/* General C language GP3 Library by Al Williams, AWC
   This library supports the GP3 I/O board (http://www.awce.com/gp3.htm)
   
   As shipped this code is meant to compile under Linux or Cygwin.
   However, if you port the following functions at the head of this
   file, the code should work for nearly any standard C compiler:

   gp3openport - Open a com port
   gp3closeport - Close the com port
   gp3write - Write a byte to the serial port
   gp3read - Read a byte from the serial port
   gp3ready - (Optional) read handshake from GP3

For more info on the other routines see the GP3 documentation
http://www.awce.com/gp3.pdf

*/

#include <stdio.h>
#include <fcntl.h>

// You may need one of these or the other
// depending on your system
//#include <termios.h>
#include <asm/termios.h>


#include "gp3lib.h"

/* Must operate at 57600 baud */
#define BAUDRATE B57600
#define DEMO 1
#define DEBUGOUT 0

#if DEBUGOUT
#define DEBUG(tag,x) printf("DEBUG: %s=%02X\n",tag,x)
#else
#define DEBUG(tag,x)
#endif

/* This is the part that is platform specific */
/* This is for Linux or Cygwin */

static int fd;  // file descriptor
static struct termios oldtio; /* terminal info */


/* Open port at 57600 baud */
int gp3openport(char *port)
{
  struct termios tio;
  int i;
  fd=open(port,O_RDWR|O_NOCTTY);
  if (fd<0)
    {
      return -1;
    }
  tcgetattr(fd,&oldtio); /* save old settings */
  memset(&tio,0,sizeof(tio));
  // Set up for 8N1, no processing */
  tio.c_cflag=BAUDRATE|CS8|CLOCAL|CREAD|CRTSCTS; 
  tio.c_iflag=IGNPAR;
  tio.c_oflag=0;
  tio.c_lflag=0;
  for (i=0;i<NCCS;i++) tio.c_cc[i]=0;
  tio.c_cc[VMIN]=1;
  tcflush(fd,TCIFLUSH);
  tcsetattr(fd,TCSANOW,&tio); // make settings take  
  return 0;
}

/* Close the port when done */
void gp3closeport()
{
  tcsetattr(fd,TCSANOW,&oldtio);
  close(fd);
}

/* Write a byte to the GP3 */
int gp3write(int byte)
{
  char c=(char)byte;
  DEBUG("write",byte);
  return write(fd,&c,1);
}

/* Read a byte from the GP3 */
/* User may call this after repeat */
unsigned int gp3read(void)
{
  char c;
  unsigned int rv;
  read(fd,&c,1);
  rv=((unsigned int)c) & 0xFF;;  
  DEBUG("read",rv);
  return rv;
}

/* Check CTS (optional if you don't want to know if GP3 is busy.
   As written, the system will handle this automatically although
   your program will block until the GP3 is available. This will
   allow you to prevent that if you implement it.
   None of the portable code uses this!

   However, if you don't implement this, you'll need to remove
   the lines that call gp3ready from the example program.
*/
int gp3ready(void)
{
  int s;
  usleep(100); // wait 100uS in case GP3 is just starting a command
  ioctl(fd,TIOCMGET,&s);
  return (s&TIOCM_CTS)==TIOCM_CTS;
}


/* This is the part that should not change for each port */

/* Private function sets argument */
static void setarg(int arg)
{
  if (arg<=255) 
    {
      gp3write(6+(arg&1));
      gp3write(0x80+(arg&1));
    }
  else 
    {
      int code=8 + (arg&1);
      gp3write(code+((arg>>8)&1));
      gp3write(0x80+(arg>>9));
      gp3write(0x80+((arg&0xFF)>>1));
    }
}



/* Function reads 16 bit result */
/* User may call this after repeat */
unsigned int gp3readword(void)
{
  unsigned int rv;
  rv=gp3read()<<8;
  return rv+gp3read();
}

/* The remaining functions match the description in the GP3 manual */

void gp3setLED(int onoff)
{
 gp3write(0xC+(onoff&1));
}

unsigned int gp3a2d(int chan)
{
  gp3write(0x30+((chan&7)<<1));
  return gp3readword();
}

void gp3high(int pin)
{
  gp3write(0x11+(pin<<1));
}

void gp3low(int pin)
{
  gp3write(0x10+(pin<<1));
}

void gp3toggle(int pin)
{
  gp3write(0x20+(pin<<1));
}

int gp3input(int pin)
{
  gp3write(0x21+(pin<<1));
  return gp3read();
}

void gp3settris(int tris)
{
  gp3write(2 + (tris & 1));
  gp3write(0x80+(tris>>1));
}

unsigned int gp3gettris(void)
{
  gp3write(0x56);
  return gp3read();
}

void gp3setpins(int byte)
{
  gp3write(4+(byte&1));
  gp3write(0x80+(byte>>1));
}

unsigned int gp3getpins(void)
{
  gp3write(0x57);
  return gp3read();
}

void gp3writeee(int address,int byte)
{
  setarg(address);
  gp3write(0x50+(byte&1));
  gp3write(0x80+(byte>>1));
}

unsigned int gp3readee(int address)
{
  setarg(address);
  gp3write(0x52);
  return gp3read();
}

void gp3pwm(int chan,int dc,int dfreq)
{
  setarg(dfreq);
  gp3write((chan==-1)?0x54:(0x60+(chan<<1))+(dc&1));
  gp3write(0x80+(dc>>1));
}

unsigned int gp3count(int pin,unsigned int ms)
{
  setarg(ms);
  gp3write(0x40+(pin<<1));
  return gp3readword();
}

unsigned int gp3rctime(int pin, int state)
{
  gp3write(0xE);
  gp3write(0xA0+(state&1)+(pin<<1));
  return gp3readword();
}

unsigned int gp3pulsein(int pin, int state)
{
  gp3write(0x70+(state&1)+(pin<<1));
  return gp3readword();
}

void gp3pulseout(int pin, int duration)
{
  setarg(duration);
  gp3write(0x41+(pin<<1));
}

void gp3freq(int pin, int ms, int freq)
{
  setarg(ms);
  gp3write(0xE);
  gp3write(0x80+(freq&1)+(pin<<1));
  gp3write(0x80+(freq>>8));
  gp3write(0x80+((freq&0xFF)>>1));
}

void gp3setcounter(int prescale,int ext)
{
  int ps=0;
  switch (prescale)
    {
    case 1: ps=0; break;
    case 2: ps=2; break;
    case 4: ps=4; break;
    case 8: ps=6; break;
    }
  gp3write(0xE);
  gp3write(0x90+ps+(ext&1));
}

unsigned int gp3counter(void)
{
  gp3write(0x5a);
  return gp3readword();
}

void gp3resetall(void)
{
  gp3write(0x5c);
}

void gp3reset(void)
{
  gp3write(0);
}

void gp3repeat(int count)
{
  setarg(count-1);
  gp3write(0x5b);
}

/* increment ARG */
void gp3argincr(void)
{
  gp3write(0x59);
}

/* decrement ARG */
void gp3argdecr(void)
{
  gp3write(0x58);
}

unsigned int gp3check(void)
{
  gp3write(1);
  return gp3read();  /* should be 0x33 */
}




/* A demo program */

#if DEMO

void pause()
{
  printf("Press ENTER\n");
  getchar();
}

int main(int argc, char *argv[])
{
  if (argc<2)
    {
      fprintf(stderr,"Usage: gp3lib port\n");
      return 1;
    }
  if (gp3openport(argv[1])<0)
    {
      perror(argv[1]);
      return 1;
    }
  printf("Connect an LED to the hardware PWM out line and output #7\n");
  printf("Resetting GP3...\n");
  gp3resetall();
  printf("GP3 version code=%02X\n",gp3check());
  printf("Turning onboard LED and out 7 on...\n");
  gp3setLED(1);
  gp3high(7);
  pause();
  printf("Turning LEDs off...\n");
  gp3setLED(0);
  gp3low(7);
  printf("Reading A/D: ");
  printf("%d\n",gp3a2d(0));
  printf("Setting hardware PWM to 16...\n");
  gp3pwm(-1,16,2000);
  printf("Now 50%% PWM on pin 7 for 5 seconds...\n");
  gp3pwm(7,128,5000);
  printf("gp3ready returns %d\n",gp3ready());  // should print 0
  gp3check(); // wait for done
  printf("gp3ready returns %d\n",gp3ready());  // should print 1
  printf("Closing port...\n");
  gp3closeport();
  return 0;

}

#endif
